#undef USE_SIMPSOLVER

#ifdef USE_SIMPSOLVER
#define SOLVER_SUPERCLASS SimpSolver
#include "SimpSolver.h"
#else
#define SOLVER_SUPERCLASS Solver
#include "Solver.h"
#endif

#include <ppl.hh>
#include <ostream>

namespace PPL = Parma_Polyhedra_Library;

#ifdef USE_DIRECT_PPL
class Constraint_Polyhedron {
  PPL::NNC_Polyhedron poly;

public:
  Constraint_Polyhedron(unsigned dimension) :
    poly(dimension, PPL::UNIVERSE) {
  }

  void add_constraint(PPL::Constraint& c) {
    poly.add_constraint(c);
  }

  bool is_empty() const {
    return poly.is_empty();
  }
};
#else
class Constraint_Polyhedron {
  PPL::LP_Problem problem;

public:
  Constraint_Polyhedron(unsigned dimension);

  void add_constraint(const PPL::Constraint& c);
  bool is_empty() const;
};
#endif

inline PPL::Constraint operator!(const PPL::Constraint& p) {
  assert(! p.is_equality());
  if (p.is_strict_inequality()) {
    return PPL::Linear_Expression(p) <= 0;
  } else {
    return PPL::Linear_Expression(p) < 0;
  }
}

struct Constraint_Polyhedron_stack_element {
  Constraint_Polyhedron p;
  Constraint_Polyhedron_stack_element *next;

  Constraint_Polyhedron_stack_element(Constraint_Polyhedron x, Constraint_Polyhedron_stack_element *n) : p(x), next(n) {
  }
};

struct Variables {
  PPL::Variable *v;

  Variables() {
    v = (PPL::Variable*) malloc(sizeof(PPL::Variable) * 1000);
    for(unsigned i=0; i<1000; i++) {
      new (v+i) PPL::Variable(i);
    }
  }

  PPL::Variable& get(unsigned i) {
    return v[i];
  }

};

bool linear_expression_less(const PPL::Linear_Expression& a, const PPL::Linear_Expression& b);

class LinearSMT : public SOLVER_SUPERCLASS {
  bool minimizeContradiction;
  int checkingRandomThreshold;

public:
  struct Linear_Expression_compare {
    bool operator() (const PPL::Linear_Expression& a, const PPL::Linear_Expression& b) const {
      return linear_expression_less(a, b);
    }
  };

private:
  std::map <std::pair<int, int>, Var> andMap;

  std::map <PPL::Linear_Expression, Var, Linear_Expression_compare> constraintMap;

  vec<PPL::Constraint*> linearSemantics;
  int linearDimension;
  Constraint_Polyhedron universePolyhedron, currentPolyhedron;
  Constraint_Polyhedron_stack_element *polyhedronStack;

  inline PPL::Constraint constraintValue(Lit p) {
    return sign(p) ? !*linearSemantics[var(p)] : *linearSemantics[var(p)];
  }

  inline bool hasConstraintValue(Lit p) {
    return var(p) < linearSemantics.size() && linearSemantics[var(p)] != 0;
  }

  void assertLiteral(Lit p);
  void checkForFinalContradiction();

  void newDecisionLevel ();
  void cancelUntil(int level);
  Lit exprAtom(const PPL::Linear_Expression& c);

  Var trueVar;

  void addImplication(Lit a, Lit b);
  void addDisjunction(Lit a, Lit b);

  typedef std::map <PPL::Linear_Expression, Var, Linear_Expression_compare>::iterator constraintMap_iterator;

public:
  typedef std::map <PPL::Linear_Expression, Var, Linear_Expression_compare>::const_iterator constraintMap_const_iterator;

  constraintMap_const_iterator constraintMap_begin() {
    return constraintMap.begin();
  }

  constraintMap_const_iterator constraintMap_end() {
    return constraintMap.end();
  }

  LinearSMT(int dimension);

  Lit mknot(Lit a) {
    return ~a;
  }

  Lit mkand(Lit a, Lit b);

  Lit mkor(Lit a, Lit b) {
    return ~(mkand(~a, ~b));
  }

  Lit mktrue() {
    return Lit(trueVar);
  }

  Lit mkfalse() {
    return Lit(trueVar, true);
  }

  Lit linearAtom(const PPL::Constraint& c);

  void generateTrivialTheoryConsequences();
  void extractContradiction();
  unsigned readLinearDimension() {
    return linearDimension;
  }
};

std::ostream& operator<<(std::ostream& o, lbool v);
