#include "smtlib_syntax.hh"
#include <iostream>
#include <exception>
#include <gmp.h>

PPL::Constraint operator>=(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  return e1.readExpr() * e2.readDivisor() >= e2.readExpr() * e1.readDivisor();
}

PPL::Constraint operator<(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  return e1.readExpr() * e2.readDivisor() < e2.readExpr() * e1.readDivisor();
}

PPL::Constraint operator<=(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  return e1.readExpr() * e2.readDivisor() <= e2.readExpr() * e1.readDivisor();
}

PPL::Constraint operator>(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  return e1.readExpr() * e2.readDivisor() > e2.readExpr() * e1.readDivisor();
}

PPL::Constraint operator==(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  return e1.readExpr() * e2.readDivisor() == e2.readExpr() * e1.readDivisor();
}

QLinear_Expression operator+(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  mpz_class oDivisor;
  mpz_lcm(oDivisor.get_mpz_t(), e1.readDivisor().get_mpz_t(), e2.readDivisor().get_mpz_t());
  return QLinear_Expression(e1.readExpr() * (oDivisor/e1.readDivisor()) +
			    e2.readExpr() * (oDivisor/e2.readDivisor()), oDivisor);
}

QLinear_Expression operator-(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  mpz_class oDivisor;
  mpz_lcm(oDivisor.get_mpz_t(), e1.readDivisor().get_mpz_t(), e2.readDivisor().get_mpz_t());
  return QLinear_Expression(e1.readExpr() * (oDivisor/e1.readDivisor()) -
			    e2.readExpr() * (oDivisor/e2.readDivisor()), oDivisor);
}

QLinear_Expression operator*(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  if (e1.readExpr().space_dimension() == 0) {
    return QLinear_Expression(e1.readExpr().inhomogeneous_term()* e2.readExpr(),
			     e1.readDivisor() * e2.readDivisor());
  } else if (e2.readExpr().space_dimension() == 0) {
    return QLinear_Expression(e2.readExpr().inhomogeneous_term()* e1.readExpr(),
			     e2.readDivisor() * e1.readDivisor());
  } else {
    std::cerr << "nonlinear term";
    exit(1);
  }
}

QLinear_Expression operator/(const QLinear_Expression& e1, const QLinear_Expression& e2) {
  if (e2.readExpr().space_dimension() == 0) {
    return QLinear_Expression(e2.readDivisor() * e1.readExpr(),
			     e2.readExpr().inhomogeneous_term() * e1.readDivisor());
  } else {
    std::cerr << "nonlinear term";
    exit(1);
  }
}

void Comparison::collectNumericVariables(std::set<std::string>& s) const {
  e1->collectNumericVariables(s);
  e2->collectNumericVariables(s);
}

void Comparison::collectPropositionalVariables(std::set<std::string>& s) const {
  e1->collectPropositionalVariables(s);
  e2->collectPropositionalVariables(s);
}

void PropositionalVariable::collectPropositionalVariables(std::set<std::string>& s) const {
  s.insert(name);
}

void PropositionalVariable::collectNumericVariables(std::set<std::string>& s) const {
}

void ManyExpression::collectPropositionalVariables(std::set<std::string>& s) const {
  for(const_iterator i=begin(); i!=end(); ++i) {
    (*i)->collectPropositionalVariables(s);
  }
}

void ManyExpression::collectNumericVariables(std::set<std::string>& s) const {
  for(const_iterator i=begin(); i!=end(); ++i) {
    (*i)->collectNumericVariables(s);
  }
}

void ManySubFormulas::collectPropositionalVariables(std::set<std::string>& s) const {
  for(const_iterator i=begin(); i!=end(); ++i) {
    (*i)->collectPropositionalVariables(s);
  }
}

void ManySubFormulas::collectNumericVariables(std::set<std::string>& s) const {
  for(const_iterator i=begin(); i!=end(); ++i) {
    (*i)->collectNumericVariables(s);
  }
}

int ManySubFormulas::size()  const {
  int total=1;
  for(std::vector<const Formula*>::const_iterator i = subFormulas.begin();
      i != subFormulas.end();
      ++i) {
    total += (*i)->size();
  }
  return total;
}

ManySubFormulas::~ManySubFormulas() {
  for(std::vector<const Formula*>::const_iterator i = subFormulas.begin();
      i != subFormulas.end();
      ++i) {
    delete(*i);
  }
}

void ManySubFormulas::printSubFormulas(std::ostream& o) const {
  for(const_iterator i = begin(); i != end(); ++i) {
    o << " " << **i;
  }
}

void And::print(std::ostream& o) const {
  o << "(and";
  printSubFormulas(o);
  o << ")";
}

void Iff::print(std::ostream& o) const {
  o << "(iff";
  printSubFormulas(o);
  o << ")";
}

void Or::print(std::ostream& o) const {

  o << "(or";
  printSubFormulas(o);
  o << ")";
}

void PropositionalVariable::print(std::ostream& o) const {
  o << name;
}

void Not::print(std::ostream& o) const {
  o << "(not " << *f << ")";
}

mpz_class mpz_one(1), mpz_zero(0);

std::ostream& operator<<(std::ostream& o, Comparison::ComparisonType cmp)
{
  switch(cmp) {
  case Comparison::LESS:
    o << "<";
    break;
  case Comparison::GREATER:
    o << ">";
    break;
  case Comparison::LESS_EQ:
    o << "<=";
    break;
  case Comparison::GREATER_EQ:
    o << ">=";
    break;
  }
  return o;
}

void Comparison::print(std::ostream& o) const {
  o << "(" << comparisonKind << " " << *e1 << " " << *e2 << ")";
}

void FLet::collectPropositionalVariables(std::set<std::string>& s) const {
  bool wasAlreadyThere = s.count(boundVariable);
  subFormula->collectPropositionalVariables(s);
  if (!wasAlreadyThere) s.erase(boundVariable);
}

void FLet::collectNumericVariables(std::set<std::string>& s) const {
  binding -> collectNumericVariables(s);
  subFormula->collectNumericVariables(s);
}

int FLet::size() const {
  return 1 + subFormula->size() + binding -> size();
}

void FLet::print(std::ostream& o) const {
  o << "(flet (" << boundVariable << " " << *binding << ") " << *subFormula << ")";
}

ManyExpression::~ManyExpression() {
  for(const_iterator i=begin(); i!=end(); ++i) {
    delete *i;
  }
}

void ManyExpressionExpression::printCompoundExpression(std::ostream& o, std::string prefix) const {
  o << "(" << prefix;
  for(const_iterator i=begin(); i!=end(); ++i) {
    o << " " << **i;
  }
  o << ")";
}

void UnaryNeg::print(std::ostream& o) const {
  o << "(~ " << *f << ")";
}

void NumericVariable::print(std::ostream& o) const {
  o << name;
}

void IntegerConstant::print(std::ostream& o) const {
  o << value;
}

void Sum::print(std::ostream& o) const {
  printCompoundExpression(o, "+");
}

void Product::print(std::ostream& o) const {
  printCompoundExpression(o, "*");
}

void ManyExpressionFormula::printListPredicate(std::ostream& o, std::string prefix) const {
  o << "(" << prefix;
  for(const_iterator i=begin(); i!=end(); i++) {
    o << " " << **i;
  }  
  o << ")";
}

void PairwiseDistinct::print(std::ostream& o) const {
  printListPredicate(o, "distinct");
}

void Equality::print(std::ostream& o) const {
  printListPredicate(o, "=");
}
int ExprLet::size() const {
  return 1+ subFormula->size();
}

void ExprLet::collectPropositionalVariables(std::set<std::string>& s) const {
  subFormula->collectPropositionalVariables(s);
}

void ExprLet::collectNumericVariables(std::set<std::string>& s) const {
  bool wasAlreadyThere = s.count(boundVariable);
  subFormula->collectNumericVariables(s);
  if (!wasAlreadyThere) s.erase(boundVariable);
  binding->collectNumericVariables(s);
}

void ExprLet::print(std::ostream& o) const {
  o << "(let (" << boundVariable << " " << *binding << ")" << *subFormula << ")";
}

int FormulaIfThenElse::size() const {
  return condition->size() + e1->size() + e2->size();
}

FormulaIfThenElse::~FormulaIfThenElse() {
  delete condition;
  delete e1;
  delete e2;
}

void FormulaIfThenElse::collectPropositionalVariables(std::set<std::string>& s) const {
  condition -> collectPropositionalVariables(s);
  e1 -> collectPropositionalVariables(s);
  e2 -> collectPropositionalVariables(s);
}

void FormulaIfThenElse::collectNumericVariables(std::set<std::string>& s) const {
  condition -> collectNumericVariables(s);
  e1 -> collectNumericVariables(s);
  e2 -> collectNumericVariables(s);
}

void FormulaIfThenElse::print(std::ostream& o) const {
  o << "(if_then_else " << *condition << " " << *e1 << " " << *e2 << ")";
}

void NumericVariable::collectPropositionalVariables(std::set<std::string>& s) const {
}

void NumericVariable::collectNumericVariables(std::set<std::string>& s) const {
  s.insert(name);
}

void IntegerConstant::collectPropositionalVariables(std::set<std::string>& s) const {
}

void IntegerConstant::collectNumericVariables(std::set<std::string>& s) const {
}

void UnaryNeg::collectPropositionalVariables(std::set<std::string>& s) const {
  f -> collectPropositionalVariables(s);
}

void UnaryNeg::collectNumericVariables(std::set<std::string>& s) const {
  f -> collectNumericVariables(s);
}

void Not::collectPropositionalVariables(std::set<std::string>& s) const {
  f -> collectPropositionalVariables(s);
}

void Not::collectNumericVariables(std::set<std::string>& s) const {
  f -> collectNumericVariables(s);
}

void Quotient::collectPropositionalVariables(std::set<std::string>& s) const {
  e1 -> collectPropositionalVariables(s);
  e2 -> collectPropositionalVariables(s);
}

void Quotient::collectNumericVariables(std::set<std::string>& s) const {
  e1 -> collectNumericVariables(s);
  e2 -> collectNumericVariables(s);
}

void Quotient::print(std::ostream& o) const {
  o << "(/ " << *e1 << " " << *e2 << ")";
}

using namespace PPL::IO_Operators;

std::ostream& operator<<(std::ostream& o, const QLinear_Expression& e) {
  mpz_class den = e.readDivisor();
  assert(den > 0);
  if (den == 1) {
    o << e.readExpr();
  } else {
    o << "(/ (" << e.readExpr() << ") " << e.readDivisor() << ")";
  }
  return o;
}

ExpressionIfThenElse::~ExpressionIfThenElse() {
  delete condition;
  delete e1;
  delete e2;
}

void ExpressionIfThenElse::collectPropositionalVariables(std::set<std::string>& s) const {
  condition -> collectPropositionalVariables(s);
  e1 -> collectPropositionalVariables(s);
  e2 -> collectPropositionalVariables(s);
}

void ExpressionIfThenElse::collectNumericVariables(std::set<std::string>& s) const {
  //std::cout << "coincoin" << std::endl;
  condition -> collectNumericVariables(s);
  e1 -> collectNumericVariables(s);
  e2 -> collectNumericVariables(s);
  //std::cout << "creating " << variableName() << std::endl;
  s.insert(variableName());
}

void ExpressionIfThenElse::print(std::ostream& o) const {
  o << "(ite " << *condition << " " << *e1 << " " << *e2 << ")";
}

std::string ExpressionIfThenElse::variableName() const {
  std::stringstream name;
  name << "#@ite_var@" << this;
  return name.str();
}
