#include "Solver.h"
#include <ostream>
#include <vector>
#include <string>
#include <map>
#include <set>
#include <gmpxx.h>
#include <ppl.hh>

class QLinear_Expression {
  PPL::Linear_Expression expr;
  mpz_class divisor;

public:
  const PPL::Linear_Expression& readExpr() const {
    return expr;
  }
  const mpz_class& readDivisor() const {
    return divisor;
  }

  QLinear_Expression(const PPL::Linear_Expression& e, const mpz_class& d) :
    expr(e), divisor(d) {
    assert(d > 0);
  }

  explicit QLinear_Expression(const PPL::Linear_Expression& e) : expr(e), divisor(1) {
  }

  QLinear_Expression() : divisor(1) {
  }

  QLinear_Expression(const mpz_class& v) : expr(v), divisor(1) {
  }
};

PPL::Constraint operator>=(const QLinear_Expression& e1, const QLinear_Expression& e2);
PPL::Constraint operator<=(const QLinear_Expression& e1, const QLinear_Expression& e2);
PPL::Constraint operator>(const QLinear_Expression& e1, const QLinear_Expression& e2);
PPL::Constraint operator<(const QLinear_Expression& e1, const QLinear_Expression& e2);
PPL::Constraint operator==(const QLinear_Expression& e1, const QLinear_Expression& e2);
QLinear_Expression operator+(const QLinear_Expression& e1, const QLinear_Expression& e2);
QLinear_Expression operator-(const QLinear_Expression& e1, const QLinear_Expression& e2);
QLinear_Expression operator*(const QLinear_Expression& e1, const QLinear_Expression& e2);
QLinear_Expression operator/(const QLinear_Expression& e1, const QLinear_Expression& e2);

std::ostream& operator<<(std::ostream& o, const QLinear_Expression& e);

inline QLinear_Expression& operator+=(QLinear_Expression& e1, const QLinear_Expression& e2) {
  e1 = e1 + e2;
  return e1;
}

inline QLinear_Expression& operator*=(QLinear_Expression& e1, const QLinear_Expression& e2) {
  e1 = e1 * e2;
  return e1;
}

inline QLinear_Expression operator*(const QLinear_Expression& e1, mpz_class x) {
  return QLinear_Expression(e1.readExpr()*x, e1.readDivisor());
}

inline QLinear_Expression operator-(const QLinear_Expression& e) {
  return QLinear_Expression(-e.readExpr(), e.readDivisor());
}

struct FormulaCompiler {
  virtual Lit mktrue() = 0;
  virtual Lit mkfalse() = 0;
  virtual Lit mkand(Lit a, Lit b) = 0;
  virtual Lit mkor(Lit a, Lit b) = 0;
  virtual Lit mkiff(Lit a, Lit b) = 0;
  virtual Lit mknot(Lit a) = 0;
  virtual Lit getPropositionalVariable(const std::string& var) const = 0;
  virtual Lit getLinearConstraintVariable(const PPL::Constraint& expr) = 0;
  virtual void assertAuxiliary(Lit a) = 0;

  virtual unsigned getNumericVariable(const std::string& var) const = 0;
  virtual ~FormulaCompiler() {
  }
  virtual void assignFormulaBinding(const std::string& var, Lit x) = 0;
  virtual void removeFormulaBinding(const std::string& var) = 0;
  virtual bool isFormulaVariableBound(const std::string& var) const = 0;
  virtual Lit readFormulaBinding(const std::string& var) const = 0;

  virtual void assignExprBinding(const std::string& var, QLinear_Expression e) =0;
  virtual void removeExprBinding(const std::string& var) =0;
  virtual bool isExprVariableBound(const std::string& var) const =0;
  virtual QLinear_Expression readExprBinding(const std::string& var) const =0;
};

class Formula {
 public:
  virtual void print(std::ostream& o) const = 0;

  virtual ~Formula() {
  };

  virtual int size() const {
    return 1;
  }

  virtual void collectPropositionalVariables(std::set<std::string>& s) const =0;
  virtual void collectNumericVariables(std::set<std::string>& s) const =0;
  virtual Lit compile(FormulaCompiler& compiler) const = 0;
};

inline std::ostream& operator<<(std::ostream& o, const Formula& f) {
  f.print(o);
  return o;
}

class FLet : public Formula {
  std::string boundVariable;
  Formula *binding, *subFormula;

public:
  FLet(std::string v, Formula *b, Formula *f) :
    boundVariable(v), binding(b), subFormula(f) {
  }

  ~FLet() {
    delete binding;
    delete subFormula;
  }

  int size() const;  
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
  Lit compile(FormulaCompiler& compiler) const;
  void print(std::ostream&) const;
};

class ManySubFormulas : public Formula {
  std::vector <const Formula*> subFormulas;

public:
  typedef std::vector <const Formula*>::const_iterator const_iterator;
  const_iterator begin() const {
    return subFormulas.begin();
  }
  const_iterator end() const {
    return subFormulas.end();
  }

  void addFormula(const Formula *f) {
    subFormulas.push_back(f);
  }

  ManySubFormulas() : subFormulas() {
  }

  ManySubFormulas(const Formula* a, const Formula* b) : subFormulas() {
    addFormula(a);
    addFormula(b);
  }

  ~ManySubFormulas();

  void printSubFormulas(std::ostream& o) const;

  int size() const;  
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
};

class And : public ManySubFormulas {
public:
  void print(std::ostream& o) const;

public:
  And() : ManySubFormulas() {
  }

  And(const Formula *a, const Formula *b) : ManySubFormulas(a, b) {
  }

  Lit compile(FormulaCompiler& compiler) const;
};


class Iff : public ManySubFormulas {
public:
  void print(std::ostream& o) const;

public:
  Iff() : ManySubFormulas() {
  }

  Iff(const Formula *a, const Formula *b) : ManySubFormulas(a, b) {
  }

  Lit compile(FormulaCompiler& compiler) const;
};

class Or : public ManySubFormulas {
public:
  void print(std::ostream& o) const;

public:
  Or() : ManySubFormulas() {
  }

  Or(const Formula *a, const Formula *b) : ManySubFormulas(a, b) {
  }

  Lit compile(FormulaCompiler& compiler) const;
};

class PropositionalVariable : public Formula {
  std::string name;

public:
  void print(std::ostream& o) const;

  PropositionalVariable(std::string s) {
    name = s;
  }

  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
  Lit compile(FormulaCompiler& compiler) const;
};

class Not : public Formula {
  Formula *f;

public:
  void print(std::ostream& o) const;

  Not(Formula *p) {
    f = p;
  }

  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
  Lit compile(FormulaCompiler& compiler) const;
};

extern mpz_class mpz_one, mpz_zero;

class Expression {
public:
  virtual void print(std::ostream& o) const = 0;

  virtual ~Expression() {
  }

  virtual int size() const {
    return 1;
  }

  virtual void collectPropositionalVariables(std::set<std::string>& s) const =0;  virtual void collectNumericVariables(std::set<std::string>& s) const =0;
  virtual QLinear_Expression compile(FormulaCompiler& compiler) const = 0;
};

class ManyExpression {
  std::vector<Expression*> expressions;

public:
  typedef std::vector<Expression*>::const_iterator const_iterator;
  
  const_iterator begin() const {
    return expressions.begin();
  }

  const_iterator end() const {
    return expressions.end();
  }

  void addExpression(Expression *e) {
    expressions.push_back(e);
  }

  ManyExpression() {
  }

  ManyExpression(Expression *e1, Expression *e2) {
    addExpression(e1);
    addExpression(e2);
  }

  ~ManyExpression();

  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
};

class ManyExpressionExpression : public Expression, public ManyExpression {
  std::vector<Expression*> expressions;

public:
  ~ManyExpressionExpression() {
  }

  ManyExpressionExpression() {
  }

  ManyExpressionExpression(Expression *e1, Expression *e2) :
    ManyExpression(e1, e2) {
  }

  void printCompoundExpression(std::ostream& o, std::string prefix) const;

  void collectPropositionalVariables(std::set<std::string>& s) const {
    ManyExpression::collectPropositionalVariables(s);
  }

  void collectNumericVariables(std::set<std::string>& s) const {
    ManyExpression::collectNumericVariables(s);
  }
};

class UnaryNeg : public Expression {
  Expression *f;

public:
  UnaryNeg(Expression *g) : f(g) {
  }

  ~UnaryNeg() {
    delete f;
  }

  void print(std::ostream& o) const;
  QLinear_Expression compile(FormulaCompiler& compiler) const;
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
};

class NumericVariable : public Expression {
  std::string name;

public:
  NumericVariable(std::string s) : name(s) {
  }  

  void print(std::ostream& o) const;
  QLinear_Expression compile(FormulaCompiler& compiler) const;
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
};

class IntegerConstant : public Expression {
  mpz_class value;

public:
  IntegerConstant(mpz_class v) : value(v) {
  }

  void print(std::ostream& o) const;
  QLinear_Expression compile(FormulaCompiler& compiler) const;
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
};

inline std::ostream& operator<<(std::ostream& o, const Expression& l) {
  l.print(o);
  return o;
}

class Sum : public ManyExpressionExpression {
public:
  Sum() {
  }

  Sum(Expression *e1, Expression *e2) : ManyExpressionExpression(e1, e2) {
  }

  void print(std::ostream& o) const;  
  QLinear_Expression compile(FormulaCompiler& compiler) const;
};

class Product : public ManyExpressionExpression {
public:
  Product() {
  }

  Product(Expression *e1, Expression *e2) : ManyExpressionExpression(e1, e2) {
  }

  void print(std::ostream& o) const;  
  QLinear_Expression compile(FormulaCompiler& compiler) const;
};

class Quotient : public Expression {
  Expression *e1, *e2;

public:
  ~Quotient() {
  }

  Quotient(Expression *f1, Expression *f2) : e1(f1), e2(f2) {
  }

  void print(std::ostream& o) const;  
  QLinear_Expression compile(FormulaCompiler& compiler) const;
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
};

class ManyExpressionFormula : public Formula, public ManyExpression {
public:
  ManyExpressionFormula() {
  }

  ManyExpressionFormula(Expression *a, Expression *b) : ManyExpression(a, b) {
  }

  int size() const {
    return 1;
  }

  void printListPredicate(std::ostream& o, std::string prefix) const;
  void collectPropositionalVariables(std::set<std::string>& s) const {
    ManyExpression::collectPropositionalVariables(s);
  }
  void collectNumericVariables(std::set<std::string>& s) const {
    ManyExpression::collectNumericVariables(s);
  }
};

class Equality : public ManyExpressionFormula {
public:
  Equality() {
  }

  Equality(Expression *a, Expression *b) : ManyExpressionFormula(a, b) {
  }

  Lit compile(FormulaCompiler& compiler) const;
  void print(std::ostream& o) const;
};

class PairwiseDistinct : public ManyExpressionFormula {
public:
  PairwiseDistinct() {
  }

  PairwiseDistinct(Expression *a, Expression *b) : ManyExpressionFormula(a, b) {
  }

  Lit compile(FormulaCompiler& compiler) const;
  void print(std::ostream& o) const;
};

class Comparison : public Formula {
  Expression *e1, *e2;

public:  
  enum ComparisonType {
    LESS_EQ, LESS, GREATER_EQ, GREATER };

private:
  enum ComparisonType comparisonKind;

public:
  void print(std::ostream& o) const;

  ComparisonType readComparisonKind() {
    return comparisonKind;
  }

  Comparison(Expression *a1, ComparisonType cmp, Expression *a2)
  : e1(a1), e2(a2), comparisonKind(cmp) {
  }

  void collectNumericVariables(std::set<std::string>& s) const;
  void collectPropositionalVariables(std::set<std::string>& s) const;

  Lit compile(FormulaCompiler& compiler) const;
};

std::ostream& operator<<(std::ostream& o, Comparison::ComparisonType cmp);

class ExprLet : public Formula {
  std::string boundVariable;
  Expression *binding;
  Formula *subFormula;

public:
  ExprLet(std::string v, Expression *e, Formula *f) :
    boundVariable(v), binding(e), subFormula(f) {
  }

  ~ExprLet() {
    delete binding;
    delete subFormula;
  }

  int size() const;
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
  Lit compile(FormulaCompiler& compiler) const;
  void print(std::ostream&) const;
};

class FormulaIfThenElse : public Formula {
  Formula *condition;
  Formula *e1, *e2;

public:
  FormulaIfThenElse(Formula *c, Formula *x1, Formula *x2) :
    condition(c), e1(x1), e2(x2) {
  }

  int size() const;
  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
  Lit compile(FormulaCompiler& compiler) const;
  void print(std::ostream&) const;

  ~FormulaIfThenElse();
};

class ExpressionIfThenElse : public Expression {
  Formula *condition;
  Expression *e1, *e2;

  std::string variableName() const;

public:
  ExpressionIfThenElse(Formula *c, Expression *x1, Expression *x2) :
    condition(c), e1(x1), e2(x2) {
  }

  void collectPropositionalVariables(std::set<std::string>& s) const;
  void collectNumericVariables(std::set<std::string>& s) const;
  QLinear_Expression compile(FormulaCompiler& compiler) const;
  void print(std::ostream&) const;

  ~ExpressionIfThenElse();
};
